# hbutils.model's package info
_PACKAGE_NAME = 'hbutils.model'
_PACKAGE_HOMEPAGE = 'https://hansbug.github.io/hbutils/main/api_doc/model/index.html'
_PACKAGE_RST = f'`{_PACKAGE_NAME} <{_PACKAGE_HOMEPAGE}>`_'
