"""
Overview:
    Collection module, include the basic utilities with collections like dict, list and tuple.
"""
from .dimension import *
from .functional import *
from .recover import *
from .sequence import *
from .stacked import *
from .structural import *
